/*
 * FILE NAME   : T1.C
 * PURPOSE     : Calculate minimal way in the labyrinth
 * PROGRAMMER  : Nikolay Dodonov.
 * LAST UPDATE : 07.10.1999
 * NOTE        : None.
 *
 * No part of this file may be changed without agreement of
 * Dodonov Nikolay (1057/2)
 */
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
 * Some constants and types
 */
#define TRUE  1
#define FALSE 0

/* Maximal distance beginning setting */
#define END_WAY 0xFFFF
/* Wall end setting */
#define WALL    0xFFFE
/* Path end setting */
#define PATH    0xFFFA
/* Beginnig value of the path in the labyrinth */
#define BEG_PATH 1
/* Beginnig value of the wall in the labyrinth */
#define BEG_WALL 0

/* Labyrinth point value function */
#define LAB_POINT( X, Y ) (*(unsigned short *)(Lab + (Y) * (W) + (X)))

/* Boolean type */
typedef short int BOOL;

/* Type of the data of on element of the list */
typedef struct tagDATA
{
  int X, Y;           /* Coordinates */
  unsigned int Deep;  /* Distance from the end of way to
                       * the current point */
} DATA;

/* The list */
typedef struct tagLIST
{
  DATA Data;
  struct tagLIST *Next;
} LIST;

/* The list */
LIST *List = NULL;
/* The labyrinth */
unsigned short *Lab = NULL;
/* Width and height of the labyrinth */
unsigned int W, H;

/***
 * Work with the list as with the turn (Put, Get)
 ***/

/* Add to the list function
 * ARGUMENTS:
 *   - pointer to the pointer to the list (first element of the list):
 *       LIST **LIST;
 *   - pointer to the data to save:
 *       DATA *Data;
 * RETURNS:
 *   Successful or not end offunction.
 * EXTERNS: None.
 */
BOOL PutToList( LIST **List, DATA *Data )
{
  /*
   * List = pointer of the element which have the next as NULL
   */
  while (*List != NULL)
    List = &((*List)->Next);

  /* Alloc memory */
  if ((*List = malloc(sizeof(LIST))) == NULL)
    /* Exit (not enough memory) */
    return FALSE;

  /* Save the data */
  memcpy(&(*List)->Data, Data, sizeof(DATA));

  /* Set pointer to the next as NULL*/
  (*List)->Next = NULL;
  return TRUE;
} /* End of 'PutToList' function */

/* Getting and deleting the data from the list function
 * ARGUMENTS:
 *   - pointer to the pointer to the list (first element of the list):
 *       LIST **LIST;
 *   - pointer to the data (return):
 *       DATA *Data;
 * RETURNS:
 *   Successful or not end offunction.
 * EXTERNS: None.
 */
BOOL GetFromList( LIST **List, DATA *Data )
{
  LIST *Old;

  /* If the list is empty - exit */
  if (*List == NULL)
    return FALSE;

  /* To the *Data write the data from the first element of the list */
  memcpy(Data, &(*List)->Data, sizeof(DATA));

  /* Save the pointer to the first element */
  Old = *List;

  /* Shift to the next pointer of the first element */
  *List = (*List)->Next;

  /* Free memory from the old first element of the list */
  free(Old);
  return TRUE;
} /* End of 'GetFromList' function */


/* Load the labyrinth from the file function.
 * ARGUMENTS:
 *   - pointer to the in file:
 *       FILE *F;
 * RETURNS:
 *   Successful.
 * EXTERNS:
 *   LAB_POINT( X, Y );
 */
BOOL LoadLab( FILE *F )
{
  int x, y;
  unsigned short V;

  /* Scan width and height of the labyrinth */
  fscanf(F, "%d %d", &W, &H);
  if ((Lab = malloc(W * H * sizeof(unsigned short))) == NULL)
    /* Not enough memory */
    return FALSE;

  /* Load walls */
  for (y = 0; y < H; y++)
    for (x = 0; x < H; x++)
    {
      fscanf(F, "%d", &V);
      if (V == BEG_PATH)
        LAB_POINT(x, y) = END_WAY;
      else
        LAB_POINT(x, y) = WALL;
    }
  /* Set begin and end of the path opened */
  LAB_POINT(0, 0) = END_WAY;
  LAB_POINT(W - 1, H - 1) = END_WAY;

  return TRUE;
} /* End of 'LoadLab' function */

/* Finding path in labyrinth function
 * ARGUMENTS: None
 * RETURNS:
 *   Successful.
 * EXTERNS:
 *   LAB_POINT( X, Y );
 */
BOOL FindPath( void )
{
  DATA d;

  /* Initialization */
  d.X = W - 1;
  d.Y = H - 1;
  d.Deep = 0;
  PutToList(&List, &d);
  while (GetFromList(&List, &d))
  {
    if (d.X < 0 || d.X >= W ||
        d.Y < 0 || d.Y >= H ||
        LAB_POINT(d.X, d.Y) == WALL ||
        LAB_POINT(d.X, d.Y) <= d.Deep)
      continue;
    /* Set the distance of path from the end of way to this point */
    LAB_POINT(d.X, d.Y) = d.Deep;
    /* If we on in the begin point */
    if (d.X == 0 && d.Y == 0)
    {
      /* Erase the list */
      while (GetFromList(&List, &d))
        ;
      return TRUE;
    }
    d.Deep++;
    d.X++;
    PutToList(&List, &d);
    d.X -= 2;
    PutToList(&List, &d);
    d.X++;
    d.Y++;
    PutToList(&List, &d);
    d.Y -= 2;
    PutToList(&List, &d);
  }
  return FALSE;
} /* End of 'FindPath' function */

/* Output results function
 * ARGUMENTS:
 *   - output file pointer:
 *       FILE *F;
 * RETURNS: None.
 * EXTERNS:
 *   LAB_POINT( X, Y );
 */
void OutputResults( FILE *F )
{
  int ST, i, x, y;
  char pl;

  x = y = 0;
  ST = LAB_POINT(0, 0);
  fprintf(F, "The way in this labyrinth STEPS = %d:\n"
      "\tOne of the way:\n", ST);
  for (i = ST; i >= 0; i--)
  {
    if (LAB_POINT(x, y) == i)
      LAB_POINT(x, y) = PATH;
    if (x - 1 >= 0 && LAB_POINT(x - 1, y) == i - 1)
    {
      x--;
      continue;
    }
    if (x + 1 < W && LAB_POINT(x + 1, y) == i - 1)
    {
      x++;
      continue;
    }
    if (y - 1 >= 0 && LAB_POINT(x, y - 1) == i - 1)
    {
      y--;
      continue;
    }
    if (y + 1 < H && LAB_POINT(x, y + 1) == i - 1)
    {
      y++;
      continue;
    }
  }

  /* Output labyrinth */
  x = W + 2;
  while (x-- > 0)
    fprintf(F, "");
  fprintf(F, "\n");
  for (y = 0; y < H; y++)
  {
    fprintf(F, "");
    for (x = 0; x < W; x++)
    {
      if (LAB_POINT(x, y) == PATH)
        pl = '';
      else if (LAB_POINT(x, y) == WALL)
        pl = '';
      else
        pl = '';
      fprintf(F, "%c", pl);
    }
    fprintf(F, "");
    fprintf(F, "\n");
  }
  x = W + 2;
  while (x-- > 0)
    fprintf(F, "");
} /* End of 'OutputResults' function */


/* The main function
 * ARGUMENTS:
 *   - number of the command line values:
 *       int ArgC;
 *   - pointer to the values:
 *       char *ArgV[];
 * RETURNS: None.
 * EXTERNS:
 *   LAB_POINT( X, Y );
 *   BOOL FindPath( void );
 *   BOOL LoadLab( FILE *F );
 *   void OutputResults( FILE *F );
 */
void main( int ArgC, char *ArgV[] )
{
  char FileNameIn[256] = "in.txt";   /* In file name */
  char FileNameOut[256] = "out.txt"; /* Out file name */
  FILE *Fin = NULL;
  FILE *Fout = NULL;
  time_t timer;
  struct tm *tblock;

  /* Gets time of day */
  timer = time(NULL);
  /* Converts date/time to a structure */
  tblock = localtime(&timer);
  /* Out program information */
  printf("\nUtility for calculation the minimal way in the labyrinth\n"
      "Written by Nick Dodonov (1057/2)\n"
      "Compiled at: " __TIME__ " of " __DATE__ "\n"
      "Now is: %s\n", asctime(tblock));

  /* If values of command line is present */
  if (ArgC > 1)
  {
    strcpy(FileNameIn, ArgV[1]);  /* Set new in file name */
    if (ArgC > 2)
      strcpy(FileNameOut, ArgV[2]);  /* Set new out file name */
  }

  if ((Fin = fopen(FileNameIn, "rt")) == NULL)
  {
    printf("\nCannot open in file \"%s\"\n", FileNameIn);
    printf("\nRight usage: %s [<InFile> [<OutFile>]]\n\n", ArgV[0]);
    return;
  }
  if ((Fout = fopen(FileNameOut, "wt")) == NULL)
  {
    fclose(Fin);
    printf("\nCannot open out file \"%s\"\n", FileNameOut);
    printf("\nRight usage: %s [<InFile> [<OutFile>]]\n\n", ArgV[0]);
    return;
  }

  /* Load labyrinth to the memory function */
  if (!LoadLab(Fin))
  {
    fclose(Fin);
    fclose(Fout);
    printf("\nCannot load labyrinth\n");
    return;
  }
  /* Close in file */
  fclose(Fin);

  /* Find minimal path function */
  if (!FindPath())
  {
    fprintf(Fout, "There is no way in this labyrinth\n");
    free(Lab);
    fclose(Fout);
    return;
  }

  /* Output results function */
  OutputResults(Fout);

  /* Free memory and close files before exit */
  free(Lab);
  fclose(Fout);
} /* End of 'main' function */

/* END OF 'T1.C' FILE */